/*****************************************************************

  mod_xslt 1.5
  $Id: mod_xslt.c,v 1.28 2001/10/21 20:47:43 pkot Exp $
                         _               _ _
     _ __ ___   ___   __| |    _  ______| | |_ 
    | '_ ` _ \ / _ \ / _` |   \ \/ / ___| | __/
    | | | | | | (_) | (_| |    >  <___  | | |_
    |_| |_| |_|\___/ \__,_|___/_/\_\____|_|___\
                         |_____|

  [ Inspired by mod_rewrite :-) ]

  This Apache Module is for translating XML using XSL.

  Copyright (c) 2000 UserActive, Inc.
  By Josh M. Nutzman <modxslt@modxslt.userworld.com>
  And Aidas Kasparas <kaspar@soften.ktu.lt>
  Copyright (c) 2001 LinuxNews.pl
  By Pawe Kot <pkot@linuxnews.pl>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; version 2.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  You may copy and distribute this code as long as this copyright
  and disclaimer remains intact.

  NO support is avaliable for this software. Should you like to
  make comments or add to the code, please e-mail
  modxslt@modxslt.userworld.com.
 
*****************************************************************/

#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "ap_config.h"
#include "util_uri.h"
#include "http_core.h"
#include "sablot.h"
#include "http_log.h"
#include "http_request.h"
#include "config.h"

#define MAXCHAR 2048 
#define MODXSLTTMPPATHLEN 100

typedef struct {
	int debug;
	int tmppath_isset;
	char tmppath[MODXSLTTMPPATHLEN];
	char xsltroot[MODXSLTTMPPATHLEN];
	char xmlroot[MODXSLTTMPPATHLEN];
} mod_xslt_cfg;

module MODULE_VAR_EXPORT mod_xslt_module;

const char *mod_xslt_debug_cmd(cmd_parms *cmd, void *m, int flag);
const char *mod_xslt_tmppath_cmd(cmd_parms *cmd, void *m, char *path);
const char *mod_xslt_xsltroot_cmd(cmd_parms *cmd, void *m, char *path);
const char *mod_xslt_xmlroot_cmd(cmd_parms *cmd, void *m, char *path);

command_rec mod_xslt_cmds[] = {
	{ "XSLTDebug", mod_xslt_debug_cmd, NULL, OR_ALL, FLAG,
	  "Is debugging of transformations allowed"},
	{ "XSLTRoot", mod_xslt_xsltroot_cmd, NULL, RSRC_CONF | ACCESS_CONF, TAKE1,
	  "Path to directory where XSLT files are stored"},
	{ "XMLRoot", mod_xslt_xmlroot_cmd, NULL, RSRC_CONF | ACCESS_CONF, TAKE1,
	  "Path to directory where XML files are stored"},
	{ NULL }
};


const char *mod_xslt_debug_cmd(cmd_parms *cmd, void *m, int flag)
{
	((mod_xslt_cfg*)m)->debug = flag;
	return NULL;
};

const char *mod_xslt_xsltroot_cmd(cmd_parms *cmd, void *m, char *path)
{
	struct stat stat;

	if (*path != '/') {
		return "XLSTRoot: directory name should be absolute - start with /\n";
	}
	if (lstat(path, &stat)) {
		return "XLSTRoot: can't get file's status\n";
	}
	if (strlen(path) >= MODXSLTTMPPATHLEN) {
		return "XSLTRoot: directory name length is too big\n";
	}

	/* Everything is OK */
	strcpy(((mod_xslt_cfg*)m)->xsltroot, path);
	return NULL;
};

const char *mod_xslt_xmlroot_cmd(cmd_parms *cmd, void *m, char *path)
{
	struct stat stat;

	if (*path != '/') {
		return "XMLRoot: directory name should be absolute - start with /\n";
	}
	if (lstat(path, &stat)) {
		return "XMLRoot: can't get file's status\n";
	}
	if (strlen(path) >= MODXSLTTMPPATHLEN) {
		return "XMLRoot: directory name length is too big\n";
	}

	/* Everything is OK */
	strcpy(((mod_xslt_cfg*)m)->xmlroot, path);
	return NULL;
};

char *mod_xslt_mkcfg(pool *p)
{
	mod_xslt_cfg *cfg = ap_palloc(p, sizeof(mod_xslt_cfg));

	cfg->debug = 0;
	cfg->xmlroot[0] = 0;
	cfg->xsltroot[0] = 0;
	return (char *)cfg;
};

static void *mod_xslt_create_server_cfg(pool *p, server_rec *s)
{
	return mod_xslt_mkcfg(p);
};

static void *mod_xslt_create_dir_cfg(pool *p, char *s)
{
	return mod_xslt_mkcfg(p);
};

int mod_xslt_save_html(char *filename, char *str)
{
	FILE *f;

	if (f = fopen(filename, "w")) {
		if (fputs(str, f) == EOF) return 99;
		fclose(f);
		return OK;
	} else return 99;
};

int mod_xslt_html_out(request_rec *r)
{
	FILE *f;
	char text[MAXCHAR];

	ap_send_http_header(r);
	f = fopen(r->filename, "r");
	if (!f) return 99;
	while (fgets(text, MAXCHAR, f)) {
		ap_rputs(text, r);
	}
	fclose(f);
	return OK;
};

int mod_xslt_debug_out(request_rec *r, char* str)
{
	r->content_type = ap_pstrcat(r->pool, "text/plain", NULL);
	ap_send_http_header(r);
	ap_rputs(str, r);
	return OK;
};

int mod_xslt_debug_err(request_rec *r, char* str)
{
	r->status = 500;
	r->content_type = ap_pstrcat(r->pool, "text/plain", NULL);
	ap_send_http_header(r);
	ap_rputs(str, r);
	return OK;
};

/* Sablot Message Handlers */

MH_ERROR mod_xslt_makecode( void *u, SablotHandle p,
			    int severity, unsigned short f, unsigned short code)
{
	return 100;
};

MH_ERROR mod_xslt_log(    void *userData, SablotHandle processor_,
			  MH_ERROR code, MH_LEVEL level, char **fields)
{
	request_rec *r = (request_rec *)userData;
	char *c = "";
	char **cc;
	return 0;
	/* This is not used at the moment... But who knows, maybe it will be
     usefull --AK */
	c = ap_psprintf(r->pool, "XSLT LOG: %d\n", level);
	for (cc = fields; *cc != NULL; cc++){
		c = ap_pstrcat(r->pool, c, *cc, "\n", NULL);
	}
	mod_xslt_debug_out(r, c);
	return 10; 
};

MH_ERROR mod_xslt_error(void *userData, SablotHandle processor_,
			MH_ERROR code, MH_LEVEL level, char **fields)
{
	request_rec *r = (request_rec *)userData;
	char *c = "";
	char **cc;

	c = ap_psprintf(r->pool, "XSLT ERROR: %d\n", level);
	for (cc = fields; *cc != NULL; cc++){
		c = ap_pstrcat(r->pool, c, *cc, "\n", NULL);
	}
	mod_xslt_debug_out(r, c);
	return 10; 
};

MessageHandler sablotMH = {
	mod_xslt_makecode,
	mod_xslt_log,
	mod_xslt_error
};

int transform(request_rec *r, char *styleSheetStr, char *inputStr, char **resultStr, char **params, char *xmlroot)
{
	int se, i = 0;
	char *argums[] = {
		"/_stylesheet", styleSheetStr,
		"/_xmlinput", inputStr,
		"/_output", NULL, //was: *resultStr,
		NULL
	};
	void *theproc;
	char *wd = NULL;

	mod_xslt_cfg *cfg = (mod_xslt_cfg*) 
		ap_get_module_config(r->per_dir_config, &mod_xslt_module);

	wd = ap_pstrcat(r->pool, "file://", xmlroot, NULL);
	se = SablotCreateProcessor (&theproc);
	if (cfg->debug)
		se |= SablotRegHandler(theproc, HLR_MESSAGE, &sablotMH, (void*)r);
	se |= SablotSetBase(theproc, wd);
	se |= SablotRunProcessor(theproc, "arg:/_stylesheet", "arg:/_xmlinput", "arg:/_output", params, argums);
	se |= SablotGetResultArg(theproc, "arg:/_output", resultStr);
	se |= SablotDestroyProcessor(theproc);
	return se;
}

#define MODULENAME PACKAGE"/"VERSION

void mod_xslt_init()
{
	ap_add_version_component(MODULENAME);
}

#define RECURSELEVEL "X-mod_xslt-recurse-level"
#define MAX_ARGS 10
#define MAX_RECURSE_LEVEL 5

#define E   ""
#define SL  "/"
#define SLP "/_"

static int mod_xslt_handler(request_rec *r)
{
	FILE *sfh;
	FILE *xfh;
	char *mimetype = NULL;
	char *tmp = NULL;
	char *tmp2 = NULL; 
	char *key = NULL;
	char *xmltext = NULL;
	char *xsltext = NULL;
	char temp[MAXCHAR + 1];
	char *xslt = NULL;
	char *recurse_level = NULL;
	int SSI = 0;
	int i = 0;
	struct stat stat1, stat2, stat3;
	int xmlsize, xslsize, noxml, noxsl, len = 0;
	char *SUFFIX = NULL, *XML_NAME = NULL, *XSLT_NAME = NULL;
	char *COMMON_PATH = NULL, *XML_PATH = NULL, *FULL_NAME = NULL;
	char *CURRENT_PATH = NULL, *XMLFILE = NULL, *XSLTFILE = NULL;
	char *HTMLFILE = NULL, *BASEXML = NULL, *BASEXSLT = NULL;
	char *SSIFILE = NULL, *BASESSI = NULL, *OUTPUTFILE = NULL;
	char *ARGS[MAX_ARGS], *xsltargs, *aux;
	FILE *f;
	int state = 0;
	mod_xslt_cfg *cfg = (mod_xslt_cfg*) 
		ap_get_module_config(r->per_dir_config, &mod_xslt_module);

	int rc = OK;
	int se = 1;
	int defxslt = 0;

	/* XMLRoot and XSLTRoot _have_ to be set */
	if (!cfg->xmlroot || !cfg->xsltroot) return HTTP_INTERNAL_SERVER_ERROR;

	/* Split URI into smaller chunks - see names.txt */
	tmp = ap_pstrcat(r->pool, r->uri, NULL);
	while (*tmp && (key = ap_getword_nulls_nc(r->pool, &tmp, '/') )) {
		if (!strcmp(key, "..")) {
			return NOT_FOUND; /* Don't allow to get out of DocumentRoot */
		}
		if (tmp2 && strlen(tmp2)) {
			if (CURRENT_PATH && strlen(CURRENT_PATH)) CURRENT_PATH = ap_pstrcat(r->pool, CURRENT_PATH, "/", tmp2, NULL);
			else CURRENT_PATH = ap_pstrcat(r->pool, tmp2, NULL);
		}
		if ((state < 2) && (key[0] == '_')) {
			key++;
			if (!state) {
				if (key[0] == '_') {
					key++;
					state++;
					XSLT_NAME = ap_pstrcat(r->pool, key, NULL);
					XML_PATH = CURRENT_PATH;
				} else {
					COMMON_PATH = CURRENT_PATH;
				}
				CURRENT_PATH = NULL;
			} else {
				XSLT_NAME = ap_pstrcat(r->pool, key, NULL);
				XML_PATH = CURRENT_PATH;
				CURRENT_PATH = NULL;
			}
			state++;
		}
		tmp2 = ap_pstrcat(r->pool, key, NULL);
	}
	if (!XML_PATH &&
	    !COMMON_PATH &&
	    CURRENT_PATH &&
	    !state) COMMON_PATH = CURRENT_PATH;
	key = ap_getword_nulls_nc(r->pool, &tmp2, '.');
	XML_NAME = ap_pstrcat(r->pool, key, NULL);
	key = ap_getword_nulls_nc(r->pool, &tmp2, '.');
	SUFFIX = ap_pstrcat(r->pool, key, NULL);
	if (!XSLT_NAME) {
		XSLT_NAME = ap_pstrcat(r->pool, XML_NAME, NULL);
		defxslt = 1;
	}

	/* Main variables */
	XMLFILE = ap_pstrcat(r->pool, cfg->xmlroot,
			     COMMON_PATH ? SL : E, COMMON_PATH ? COMMON_PATH : E,
			     XML_PATH ? SL : E, XML_PATH ? XML_PATH : E,
			     "/", XML_NAME, ".xml", NULL);
	XSLTFILE = ap_pstrcat(r->pool, cfg->xsltroot,
			      "/", SUFFIX,
			      COMMON_PATH ? SL : E, COMMON_PATH ? COMMON_PATH : E,
			      "/", XSLT_NAME, ".xslt", NULL);
	HTMLFILE = ap_pstrcat(r->pool, r->filename, NULL);
	BASEXML = ap_pstrcat(r->pool,
			     COMMON_PATH ? SL : E, COMMON_PATH ? COMMON_PATH : E,
			     XML_PATH ? SL : E, XML_PATH ? XML_PATH : E,
			     "/", XML_NAME, ".xml", NULL);
	BASEXSLT = ap_pstrcat(r->pool,
			      "/", SUFFIX,
			      COMMON_PATH ? SL : E, COMMON_PATH ? COMMON_PATH : E,
			      "/", XSLT_NAME, ".xslt", NULL);
	if (cfg->debug) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "COMMON_PATH: %s", COMMON_PATH);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "XML_PATH: %s", XML_PATH);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "XML_NAME: %s", XML_NAME);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "XSLT_NAME: %s", XSLT_NAME);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "SUFFIX: %s", SUFFIX);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "XML: %s", BASEXML);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "XSLT: %s", BASEXSLT);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "XML: %s", XMLFILE);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "XSLT: %s", XSLTFILE);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "HTML: %s", HTMLFILE);
	}

	/* Will the served doc be static or server parsed? Default - static */
	OUTPUTFILE = HTMLFILE;
	f = ap_pfopen(r->pool, XSLTFILE, "r");
	if (f) {
		fgets(temp, MAXCHAR, f);
		if (!strncmp("<!--SSI-->", temp, 10)) {
			SSI = 1;
			SSIFILE = ap_pstrcat(r->pool, ap_document_root(r),
					     COMMON_PATH ? SL : E, COMMON_PATH ? COMMON_PATH : E,
					     XML_PATH ? SLP: E, XML_PATH ? XML_PATH : E,
					     defxslt ? E : SLP, defxslt ? E : XSLT_NAME,
					     "/", XML_NAME, ".shtml", NULL);
			BASESSI = ap_pstrcat(r->pool,
					     COMMON_PATH ? SL : E, COMMON_PATH ? COMMON_PATH : E,
					     XML_PATH ? SLP: E, XML_PATH ? XML_PATH : E,
					     defxslt ? E : SLP, defxslt ? E : XSLT_NAME,
					     "/", XML_NAME, ".shtml", NULL);
			if (cfg->debug) {
				ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "SSI: %s", SSIFILE);
			}
			OUTPUTFILE = SSIFILE;
		}
		ap_pfclose(r->pool, f);
	}

	noxml = stat(XMLFILE, &stat2);
	noxsl = stat(XSLTFILE, &stat3);
	xmlsize = stat2.st_size;
	xslsize = stat3.st_size;
	if (!stat(OUTPUTFILE, &stat1)) {                 // HTML/SSI file does exist
		if (!r->args &&                          // do not cache dynamic docs
		    !noxml &&                            // XML file does exist
		    (stat1.st_mtime > stat2.st_mtime) && // HTML/SSI is newer then XML file
		    !noxsl &&                            // XSLT file does exist
		    (stat1.st_mtime > stat3.st_ctime)) { // HTML/SSI is newer then XSLT file
			if (cfg->debug) ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "Serving static output");
			if (SSI) {
				ap_internal_redirect(BASESSI, r);
				return OK;
			} else {
				if (mod_xslt_html_out(r) == OK) return OK;
			}
		}
		if (!stat1.st_size) {
			int timeout = 100;

			while (!stat1.st_size && timeout--) { // HTML/SSI file is empty -> someone is creating it or there was some error before
				sleep(1);                    // wait
				lstat(OUTPUTFILE, &stat1);
			}
			if (timeout)
				if (SSI) {
					ap_internal_redirect(BASESSI, r);
					return OK;
				} else {
					if (mod_xslt_html_out(r) == OK) return OK;
				}
		}
		utime(OUTPUTFILE);                     // Change st_mtime of the file - we will change it. Noone else should do it
	} else {
		FILE *f;

		f = ap_pfopen(r->pool, OUTPUTFILE, "w"); // Touch the file
		if (!f) ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "Can't open file %s.", OUTPUTFILE);
		else ap_pfclose(r->pool, f);
 	}

	mimetype = ap_pstrcat(r->pool, r->filename, NULL);

	key = NULL;

	recurse_level = (char *)ap_table_get(r->headers_in, RECURSELEVEL);
	if (recurse_level) {
		if (strlen(recurse_level) > MAX_RECURSE_LEVEL) {
			if (cfg->debug) {
				mod_xslt_debug_err(r, "XSLT DEBUG: Recursion is too deep\n"); 
				return SERVER_ERROR;
			} else {
				ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "Recursion is too deep (%d levels)", MAX_RECURSE_LEVEL);
				return SERVER_ERROR;
			}
		} else
			ap_table_set(r->headers_in, RECURSELEVEL, ap_pstrcat(r->pool, recurse_level, "1", NULL));
	} else ap_table_set(r->headers_in, RECURSELEVEL, "1");

	/* Now each has its own copy of the filename. Parse it out and get what we want. */
	while (*mimetype && (key = ap_getword_nulls_nc(r->pool, &mimetype, '/') )) {}
	while (*key && (mimetype = ap_getword_nulls_nc(r->pool, &key, '.') )) {}
  
	if (!strcmp("xml", mimetype)) {
		if (cfg->debug) {
			mod_xslt_debug_err(r, "XSLT DEBUG: \"AddHandler mod_xslt .xml\" will cause infinite recursion. Remove it!\n"); 
			return SERVER_ERROR;
		} else {
			ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "\"AddHandler mod_xslt .xml\" will cause infinite recursion. Remove it!");
			return SERVER_ERROR;
		}
	}


	/* Ok, now put together the proper file names and the content-type. */

	/* XML */
	sfh = ap_pfopen(r->pool, XMLFILE, "r");
	if (sfh == NULL) {
		unlink(OUTPUTFILE);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "File not found %s", XMLFILE);
		return NOT_FOUND;
	}
	xmltext = mmap(NULL, xmlsize, PROT_READ, MAP_PRIVATE, fileno(sfh), 0);
	if (xmltext == MAP_FAILED) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "XML mmap failed: %d, size: %d", errno, xmlsize);
	}
	ap_pfclose(r->pool, sfh);
			
	temp[0] = temp[MAXCHAR] = 0;
	xfh = ap_pfopen(r->pool, XSLTFILE, "r");
	if (xfh == NULL) {
		unlink(OUTPUTFILE);
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "File not found %s", XSLTFILE);
		return NOT_FOUND;
	}
	xsltext = mmap(NULL, xslsize, PROT_READ, MAP_PRIVATE, fileno(xfh), 0);
	if (xsltext == MAP_FAILED) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, "XSL mmap failed: %d, size: %d", errno, xslsize);
	}

	ap_pfclose(r->pool, xfh);

	/* Ugly as hell... */
	if (SSI) len = 11;
	else len = 0;

	/* Parameters to xslt */
	for (i = 0; i < MAX_ARGS; i++) ARGS[i] = 0;
	if (r->args && r->args[0]) {
		int clen = 0, alen = strlen(r->args);
		xsltargs = ap_pstrcat(r->pool, r->args, NULL);
		i = 0;

		while (aux = strsep(&xsltargs, "=")) {
			int nlen = 0;

			ARGS[i++] = aux;
			aux = strsep(&xsltargs, "&");
			if (!aux) break;
			ARGS[i++] = aux;

		}
	}
	if ((i > 0) && (i % 2)) ARGS[i-1] = 0;
	
	if (rc == OK) {
		/* This is fun! Now actually translate the documents, since we've loaded both of them. */
		if ((xsltext != MAP_FAILED) && (xmltext != MAP_FAILED)) {
			se = transform(r, xsltext+len, xmltext, &xslt, ARGS, cfg->xmlroot);
			munmap(xmltext, xmlsize);
			munmap(xsltext, xslsize);
		} else {
			if (xmltext != MAP_FAILED) {
				munmap(xmltext, xmlsize);
			}
			if (xsltext != MAP_FAILED) {
				munmap(xsltext, xslsize);
			}
			se = 1;
		}
		if (se) {
			rc = SERVER_ERROR;
		}
		if (xslt && rc == OK) {
			mod_xslt_save_html(OUTPUTFILE, xslt);
			if (SSI) {
				ap_internal_redirect(BASESSI, r);
				rc = OK;
			} else {
				ap_send_http_header(r);
				ap_rputs(xslt, r);
			}
		} else {
			if (cfg->debug) rc = OK; 
			else rc = SERVER_ERROR;
		}
		if (xslt) {
			SablotFree(xslt);
			xslt = NULL;
		}
	}
	return rc;
}

/* Dispatch list of content handlers */
static const handler_rec mod_xslt_handlers[] = { 
	{ PACKAGE, mod_xslt_handler }, 
	{ NULL, NULL }
};

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT mod_xslt_module = {
	STANDARD_MODULE_STUFF, 
	mod_xslt_init,              /* module initializer                  */
	mod_xslt_create_dir_cfg,    /* create per-dir    config structures */
	NULL,                       /* merge  per-dir    config structures */
	mod_xslt_create_server_cfg, /* create per-server config structures */
	NULL,                       /* merge  per-server config structures */
	mod_xslt_cmds,              /* table of config file commands       */
	mod_xslt_handlers,          /* [#8] MIME-typed-dispatched handlers */
	NULL,                       /* [#1] URI to filename translation    */
	NULL,                       /* [#4] validate user id from request  */
	NULL,                       /* [#5] check if the user is ok _here_ */
	NULL,                       /* [#3] check access by host address   */
	NULL,                       /* [#6] determine MIME type            */
	NULL,                       /* [#7] pre-run fixups                 */
	NULL,                       /* [#9] log a transaction              */
	NULL,                       /* [#2] header parser                  */
	NULL,                       /* child_init                          */
	NULL,                       /* child_exit                          */
	NULL                        /* [#0] post read-request              */
};
